package org.sun.jsat.trees;

import java.util.Comparator;
import java.util.List;
import javax.swing.tree.DefaultMutableTreeNode;
import java.util.Collections;
import java.util.HashMap;
import java.util.Vector;
import javax.swing.tree.TreeNode;
import securityanalyzer.resourcemodel.Definition;
import securityanalyzer.resourcemodel.FilledParameterContainer;

/**
 *
 * 
 */
public class RefreshableTreeNode<BaseType> extends DefaultMutableTreeNode {

    /** set this to true when a generation process starts to avoid cyclic generation */
    protected boolean childrenGenerated = false;
    private String name;
    protected HashMap<Object, RefreshableTreeNode> childrenMap;
    protected Object updateMutex = new Object();
    protected BaseType baseObject;
    private boolean highlight = false;

    public RefreshableTreeNode(BaseType baseObject, String name) {
        this.name = name;
        this.baseObject = baseObject;
    }

    @Override
    public boolean isLeaf() {
        if (!childrenGenerated) {
            generateAllChildren();
        }
        return super.isLeaf();
    }

    @Override
    public int getChildCount() {
        if (!childrenGenerated) {
            generateAllChildren();
        }
        return super.getChildCount();
    }

    public boolean childrenSortable() {
        if (children == null) {
            return false;
        }
        for (Object o : children) {
            if (!(o instanceof RefreshableTreeNode)) {
                return false;
            }
        }
        return true;
    }

    @SuppressWarnings("unchecked")    //This actually is "checked" by childrenSortable()
    public void sortChildren() {
        //if (children == null) return; //already done by childrenSortable()
        if (childrenSortable()) {
            Vector<RefreshableTreeNode> list =
                    (Vector<RefreshableTreeNode>) children;
            Collections.sort(list,
                    new RefreshableTreeNodeComparator());
            //This is already done by refresh()
            /*for (Object o : children) {
                ((RefreshableTreeNode) o).sortChildren();
            }*/
        }
    }

    /**
     * generates all the children of this tree node by flushing the child list,
     * then recreating a new object for each child.
     */
    public void generateAllChildren() {
        synchronized (updateMutex) {
            generationActions();
            generateMissingChildren();
        }
    }

    /**
     * generates only the children that are missing
     */
    public void generateMissingChildren() {
        synchronized (updateMutex) {
            // I don't know what children are missing, so I'll just pass the message along
            generationActions();
            for (Object n : children) {
                if (n instanceof RefreshableTreeNode) {
                    ((RefreshableTreeNode) n).generateMissingChildren();
                }
            }
            sortChildren();
        }
    }
    
    public void removeExtraChildren() {
        //I don't know who to remove, so I won't.
        return;
    }
    
    /**
     * call this to set up the children Vector and Map, and set the generated 
     * flag
     */
    protected void generationActions() {
        synchronized (updateMutex) {
            childrenGenerated = true;
            if (children == null) {
                children = new Vector<TreeNode>();
            }
            if (childrenMap == null) {
                childrenMap = new HashMap<Object, RefreshableTreeNode>();
            }
        }
    }

    /**
     * Updates the order and status of children without destroying them. Also
     * updates any children.
     */
    public void refresh() {
        synchronized (updateMutex) {
            generateMissingChildren();
            removeExtraChildren();
            if (children == null) {
                return;
            }
            for (Object o : children) {
                if (o instanceof RefreshableTreeNode) {
                    ((RefreshableTreeNode) o).refresh();
                }
            }
            sortChildren();
        }
    }

    @Override
    public String toString() {
        if (name == null || name.equals("")) return "(blank)";
        return name;
    }
    
    /**
     * To make tool tips happen, do something like this:
     * 
     * JTree tree = new JTree(nodes[0]) {
      public String getToolTipText(MouseEvent evt) {
        if (getRowForLocation(evt.getX(), evt.getY()) == -1)
          return null;
        TreePath curPath = getPathForLocation(evt.getX(), evt.getY());
        return ((RefreshableTreeNode) curPath.getLastPathComponent())
            .getToolTipText();
      }
     * @return
     */
    public String toolTipText(){
        if (baseObject instanceof Definition){
            return ((Definition)baseObject).getDescription();
        } else if (baseObject instanceof FilledParameterContainer){
            if(((FilledParameterContainer)baseObject).getBaseObject() instanceof Definition){
                return ((Definition)((FilledParameterContainer)baseObject).getBaseObject()).getDescription();
            }
        }
        return "";
    }


    protected RefreshableTreeNode getOrAddNode(Object nodeMapping, String nodeName) {
        synchronized (updateMutex) {
            if (childrenMap == null) {
                childrenMap = new HashMap<Object, RefreshableTreeNode>();
            }
            RefreshableTreeNode node = childrenMap.get(nodeMapping);
            if (node != null) {
                return node;
            }
            node = new RefreshableTreeNode(null, nodeName);
            childrenMap.put(nodeMapping, node);
            add(node);
            return node;
        }
    }

    public void setBaseObject(BaseType base){
        this.baseObject = base;
    }
    
    public BaseType getBaseObject(){
        return baseObject;
    }
    
    public void setHighlight(boolean highlight){
        this.highlight = highlight;
    }

    boolean isHighlighted() {
        return this.highlight;
    }
    
    private class RefreshableTreeNodeComparator
            implements Comparator<RefreshableTreeNode> {

        
        public int compare(RefreshableTreeNode o1, RefreshableTreeNode o2) {
            return o1.toString().compareTo(o2.toString());
        }
    }
    
    public void setName(String name) {
        this.name = name;
    }
    
    public boolean equals(Object o) {
        if (o == null) {
            return false;
        }
        
        if(!(o instanceof RefreshableTreeNode)) {
            return false;
        }
        
        RefreshableTreeNode node = (RefreshableTreeNode) o;
        if(name == null) {
            if(node.name != null) {
                return false;
            }
        }
        else {
            if(!name.equals(node.name)) {
                return false;
            }
        }
        
        if(baseObject == null) {
            if(node.baseObject != null) {
                return false;
            }
        }
        else {
            if(!baseObject.equals(node.baseObject)) {
                return false;
            }
        }
        
        return true;
    }
}
