package securityanalyzer.analysismodel.tools.soot;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import securityanalyzer.analysismodel.AnalysisModel;
import securityanalyzer.analysismodel.CodeBlock;
import securityanalyzer.analysismodel.Method;
import securityanalyzer.analysismodel.MethodInvocation;
import securityanalyzer.analysismodel.tools.StaticAnalysisTool;
import soot.Kind;
import soot.SootClass;
import soot.SootMethod;
import soot.jimple.toolkits.callgraph.Edge;
import soot.util.Chain;

public class SootAnalysisTool extends StaticAnalysisTool {

	private String analysisType = "SootStaticAnalysis";
	String[] applicationClasses;
	String mainClass;
	String processDir;
	AnalysisModel model;
	List<SootMethod> initialSootMethods;
	List<Method> initialMethods;

	public SootAnalysisTool(String[] classes, String mainClass) {
		this.applicationClasses = classes;
		this.mainClass = mainClass;
		initialMethods = new ArrayList<Method>();
	}

	public SootAnalysisTool(String processDir, String mainClass) {
		this.applicationClasses = new String[0];
		this.processDir = processDir;
		this.mainClass = mainClass;
		initialMethods = new ArrayList<Method>();
	}

	public String getAnalysisType() {
		return this.analysisType;
	}

	/**
	 * Runs a static security analysis using soot to create a callgraph and find initial
	 * permission information
	 */
	public void runAnalysis() {

		setupModel();

		doAnalysis();

	}

	/**
	 * Uses a SootStaticModel to set up the AnalysisModel
	 */
	private void setupModel() {
		SootStaticModel ssm = new SootStaticModel();
		
		try {

			if (processDir != null) {
				ssm.setProcessDir(processDir);
			}

			for (String s : applicationClasses) {
				ssm.addClassToScene(s);
			}

			ssm.setMainClass(mainClass);

			ssm.init();
			
		} catch (Exception e) {
			System.out.println("Couldn't set up Soot.  Soot reported:  \n"
					+ e.getMessage());
			return;
		}

		ssm.getInitialMethods();

		// done initializing ssm, now add to our model
		addClassesToModel(ssm.getClasses());

		// all classes and methods are there, now need to have the edges
		// of the call graph
		addEdgesToModel(ssm.getEdges());

		//addCallGraphToModel(ssm.getCallGraph());

		ssm.clearSootData();
	}

	public void setAnalysisType(String analysisType) {
		this.analysisType = analysisType;
	}

	private void addEdgesToModel(Collection<Edge> edges) {

		for (Edge e : edges) {
			Method source = (Method) AnalysisModel.findByName(e.src()
					.getSignature()), target = (Method) AnalysisModel
					.findByName(e.tgt().getSignature());

			boolean priv = false, thread = false;

			if (e.kind().equals(Kind.PRIVILEGED)) {
				priv = true;
			}

			if (e.kind().equals(Kind.THREAD)) {
				thread = true;
			}

			addMethodInvocation(source, target, priv, thread);

		}

	}

	/**
	 * Add all classes from soot to the analysis model
	 * @param classes
	 */
	private void addClassesToModel(Chain<SootClass> classes) {

		for (SootClass c : classes) {
			CodeBlock cb = addClass(c.getJavaStyleName());
			if (c.isApplicationClass()) {
				cb.setAppCode(true);
			}
			addClassMethodsToModel(c, cb);

		}
	}

	private void addClassMethodsToModel(SootClass c, CodeBlock cb) {

		List<SootMethod> methods = c.getMethods();

		for (SootMethod sm : methods) {
			Method m = addMethod(sm.getSignature(), cb.getFullyQualifiedName());

			// If this method is tagged with permissions, it is a starting point
			// for the graph traversal
			PermissionDescriptionTag pt = (PermissionDescriptionTag) sm
					.getTag("PermissionDescriptionTag");

			if (pt != null) {
				initialMethods.add(m);
				getRequirementsFromTag(sm, m);
			}
		}

	}

	/**
	 * Adds the given Soot CallGraph to the AnalysisModel
	 * @param cg
	 */
/*	private void addCallGraphToModel(CallGraph cg) {

		for (SootMethod sm : initialSootMethods) {
			Method m = addMethod(sm.getSignature(), sm.getDeclaringClass()
					.getJavaStyleName());
			initialMethods.add(m);

			getRequirementsFromTag(sm, m);

			addByEdges(m, cg.edgesOutOf(sm), cg);

		}

	}*/

	/**
	 * This adds methods that are at the source of the given edges into the
	 * AnalysisModel as Method objects, and works recursively through the call
	 * graph. Stopping case is when a method with the given signature already
	 * exists (this indicates a cycle in the graph)
	 * 
	 * @param fromMethod
	 * @param edges
	 * @param cg
	 */
/*	private void addByEdges(Method fromMethod, Iterator<Edge> edges,
			CallGraph cg) {
		while (edges.hasNext()) {

			Edge e = edges.next();

			SootMethod sm = e.src();

			// Make sure this method hasn't been already added
			if (!hasMethod(sm.getSignature())) {

				Method m = addMethod(sm.getSignature(), sm.getDeclaringClass()
						.getJavaStyleName());

				addMethodInvocation(fromMethod, m, e.kind() == Kind.PRIVILEGED,
						e.kind() == Kind.THREAD);

				addByEdges(m, cg.edgesInto(sm), cg);
			}

		}
	}*/

	/**
	 * The Soot data analysis of Permissions stores the information in tags which contain
	 * PermissionDescriptions.  This method converts from one to the other
	 * @param sm
	 * @param m
	 */
	private void getRequirementsFromTag(SootMethod sm, Method m) {

		PermissionDescriptionTag pt = (PermissionDescriptionTag) sm
				.getTag("PermissionDescriptionTag");

		if (pt == null) {
			return;
		}

		m.addPermissionRequirements(pt.getPermissions());

	}

	/**
	 * Does the analysis by starting at initialMethods and traversing each one
	 * backwards with the graph traversal algorithm
	 */
	private void doAnalysis() {

		for (Method m : initialMethods) {
			traverseByMethod(m);
		}

	}

	/**
	 * Traverses all methods with outgoing method invocations into Method m
	 * @param m
	 */
	private void traverseByMethod(Method m) {

		for (MethodInvocation inv : m.getAllMethodInvocationsIn()) {

			if (inv.isPriv()) {
				traversePrivileged(inv.getInvokingMethod(), inv
						.getMethodCalled());
			} else if (inv.isThread()) {
				traverseThread(inv.getInvokingMethod(), inv.getMethodCalled());
			} else {
				traverseNormal(inv.getInvokingMethod(), inv.getMethodCalled());
			}
		}
	}

	/**
	 * The traversal strategy for a normal method invocation
	 * All permissions are added from callee to caller, and then the caller
	 * is traversed through all edges going out of it.  This also stops if the
	 * caller is library code and the callee is application code (library callback
	 * case)
	 * @param caller
	 * @param callee
	 */
	private void traverseNormal(Method caller, Method callee) {

		//If the caller already has all those permissions, done with this branch
		if (addPermissions(caller, callee)) {
			return;
		}

		// If the caller is library code calling into app code, stop here
		// TODO: may need to think about this one more
		if (leavingAppCode(caller, callee)) {
			return;
		}

		traverseByMethod(caller);
	}

	/**
	 * Traversal strategy for a THREAD edge.  The caller of a thread edge is
	 * the context that called Thread.start().  This caller does not require the
	 * permissions of the thread run method - the context that created it does.
	 * So, the traversal travels back to the constructor of the thread class in
	 * question, which will then traverse to all methods that created this kind
	 * of thread. (This is a bit conservative, and possible changes to this will
	 * be considered.)
	 * @param caller
	 * @param callee
	 */
	private void traverseThread(Method caller, Method callee) {
		//Here the callee is (? extends Thread).run()
		//And we get the children of the callee's parent, which will contain
		// all the methods of the thread class in question
		Set<CodeBlock> threadMethods = callee.getParent().getChildren();

		for (CodeBlock cb : threadMethods) {
			if (cb instanceof Method) {
				//If this method is a constructor...
				if (cb.getFullyQualifiedName().contains("<init>")
						&& cb instanceof Method) {

					if (!addPermissions((Method) cb, callee)) {
						traverseByMethod((Method) cb);
					}
				}
			}
		}
	}

	/**
	 * All we need to do here is to add the permissions to the method that called
	 * the doPrivileged and we are done, since this context has taken
	 * responsibility for the permissions needed lower on the call stack
	 * @param caller
	 * @param callee
	 */
	private void traversePrivileged(Method caller, Method callee) {
		addPermissions(caller, callee);
	}

	/**
	 * Returns true iff m1 is not application code and m2 is.  This is
	 * used to tell if the graph traversal is leaving application code
	 * and entering library code (this is a library callback into app code,
	 * a case that we don't want to consider yet)
	 * @param m1
	 * @param m2
	 * @return
	 */
	private boolean leavingAppCode(Method m1, Method m2) {
		return !m1.isAppCode() && m2.isAppCode();
	}

	/**
	 * Adds all permissions requirements from m2 to m1.  Returns true
	 * if m1 already had all these permissions, false otherwise.
	 * @param m1
	 * @param m2
	 * @return
	 */
	private boolean addPermissions(Method m1, Method m2) {

		if (m1.getAllPermissionRequirements().containsAll(
				m2.getAllPermissionRequirements())) {
			return true;
		}

		m1.getAllPermissionRequirements().addAll(
				m2.getAllPermissionRequirements());
		return false;

	}

}
