package securityanalyzer.resourcemodel;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * 
 * Defines an Action. An action is a child of a Resource which has a set of
 * parameters, a set of rules to apply, and mappings between the action's
 * parameters and the rules' parameters.
 * 
 */
public class ActionDefinition extends Definition implements ParameterContainer {

    private ParameterDefinition groupBy;
    /** The parent Resource of this Action */
    private ResourceDefinition resource;
    /** The sets of rules to apply to allow this action keyed by rule type */
    private Map<String, Set<RuleDefinition>> allRules;
    /**
     * Contains sets of ActionDefinitions. If all ActionDefinitions in a given
     * set are granted, this action is also granted by implication
     */
    /*private Set<Set<ActionDefinition>> impliedBy;*/
    /** Set of parameters for this Action */
    private List<ParameterDefinition> parameters;
    /**
     * A set of ParameterApplications for each PermissionDescription, defining
     * how to create a Rule's parameters
     */
    private Map<String, Set<AppliedRule>> parameterApplications;

    /**
     * Constructor
     * 
     * @param name
     *            The name of the Action
     * @param description
     *            The description of the Action
     * @param parent
     *            The Resource this action applies to
     */
    public ActionDefinition(String name, String description,
            ResourceDefinition parent) {
        this.resource = parent;
        this.name = name;
        this.description = description;
        this.allRules = new HashMap<String, Set<RuleDefinition>>();
        /*this.impliedBy = new HashSet<Set<ActionDefinition>>();*/
        this.parameters = new ArrayList<ParameterDefinition>();
        this.parameterApplications = new HashMap<String, Set<AppliedRule>>();
    }

    /**
     * Contains sets of ActionDefinitions. If all ActionDefinitions in a given
     * set are granted, this action is also granted by implication
     * 
     * @return
     */
    /*	public Set<Set<ActionDefinition>> getActionSetsThatImplyThis() {
    return impliedBy;
    }*/
    /**
     * Add a set of actions that imply this action. If all actions in this set
     * are allowed, this action is allowed by implication.
     * 
     * @param a
     *            Set of actions that would cause this action to be allowed
     * @return true if the set did not already contain 'a'
     */
    /*	public boolean addActionSetThatImpliesThis(Set<ActionDefinition> a) {
    return impliedBy.add(a);
    }*/
    /**
     * Get all the rules of the specified type required to allow this action
     * 
     * @param type The type of rule to look up
     * 
     * @return
     */
    public Set<RuleDefinition> getRequiredRules(String type) {
        return allRules.get(type);
    }

    /**
     * Add a rule to be required by this action
     * 
     * @param r
     *            rule to add
     * @return true if the set of rules did not already contain 'r'
     */
    public boolean addRequiredPermission(RuleDefinition r) {
        Set<RuleDefinition> rules = allRules.get(r.getType());
        if (rules == null) {
            rules = new HashSet<RuleDefinition>();
            allRules.put(r.getType(), rules);
        }
        return rules.add(r);
    }

    public List<ParameterDefinition> getParameters() {
        return parameters;
    }

    public ParameterDefinition getParameter(String name) {
        for (ParameterDefinition pd : parameters) {
            if (pd.name.equals(name)) {
                return pd;
            }
        }

        return null;
    }

    public boolean addParameter(ParameterDefinition p) {
        return parameters.add(p);
    }

    /**
     * Add a mapping from a RuleDefinition's parameter to a specified
     * value. This value can include variables surrounded by '{' and '}'. These
     * variables will be replaces with ActionInstance's parameter mappings.
     * 
     * @param rd
     *            The RuleDefinition containing the parameter to map
     * @param parameterName
     *            The name of the RuleDefinition's parameter to map
     * @param parameterValue
     *            The value to map the RuleDefinition's parameter to (can
     *            include variables)
     *            
     * @return true if this is a new mapping, false if it replaces a mapping
     */
    public boolean addRuleMapping(RuleDefinition rd,
            Map<String, String> nameValueMap) {

        Set<AppliedRule> youNeedToNameIt = parameterApplications.get(rd.getType());

        if (youNeedToNameIt == null) {
            youNeedToNameIt = new HashSet<AppliedRule>();
            parameterApplications.put(rd.getType(), youNeedToNameIt);
        }

        return youNeedToNameIt.add(new AppliedRule(rd, nameValueMap));
    }

    /**
     * Resolves all parameters in a RuleDefinition to their proper values
     * by replacing variables from the given parameter map.
     * 
     * @param rd
     *            the RuleDefinition to do parameter work with
     * @param parameterMappings
     *            The mappings from variable name to variable value
     * @return A map of parameter name to parameter value with all variables
     *         replaced, or null if this ActionDefinition does not contain the
     *         specified RuleDefinition
     */
    /*	public Map<String, String> resolveRuleParameters(
    RuleDefinition rd, Map<String, String> parameterMappings) {
    // check to see if we deal with that permission
    Set<AppliedRule> paramApps = parameterApplications.get(rd);
    if (paramApps == null)
    return null;
    
    Map<String, String> ret = new HashMap<String, String>();
    for (AppliedRule pa : paramApps) {
    ret.put(pa.name, resolveVariables(pa.value, parameterMappings));
    }
    return ret;
    }*/
    /**
     * Get the parent resource of this Action
     * 
     * @return
     */
    public ResourceDefinition getResource() {
        return resource;
    }

    /**
     * 
     */
    public Set<AppliedRule> getParameterApplications(String type) {
        return parameterApplications.get(type);
    }

    public ParameterDefinition getGroupBy() {
        return groupBy;
    }

    public void setGroupBy(ParameterDefinition pd) {
        this.groupBy = pd;
    }
    
    @Override
    public int hashCode() {
        String chstr = name + description + resource.getName() + resource.getDescription();
        //TODO make this stronger
        return chstr.hashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof ActionDefinition)) return false;
        return this.hashCode() == obj.hashCode();
    }

    /**
     * 
     * Defines a way to resolve a PermissionDefinition's parameter
     * 
     */
    public class AppliedRule {

        /**
         * Names of the PermissionDefinition's parameter mapped to values to
         * resolve the parameter to (can contain variables enclosed in '{' and
         * '}'
         */
        public Map<String, String> nameValueMap;
        /**
         * 
         */
        public RuleDefinition ruleDefinition;

        /**
         * Constructor
         * 
         * @param name
         *            Name of the PermissionDefinition's parameter
         * @param value
         *            Value to resolve the parameter to (can contain variables
         *            enclosed in '{' and '}'
         */
        public AppliedRule(RuleDefinition rd) {
            nameValueMap = new HashMap<String, String>();
            ruleDefinition = rd;
        }

        /**
         * Constructor
         * 
         * @param name
         *            Name of the PermissionDefinition's parameter
         * @param value
         *            Value to resolve the parameter to (can contain variables
         *            enclosed in '{' and '}'
         */
        public AppliedRule(RuleDefinition rd, Map<String, String> nameValueMap) {
            this.nameValueMap = nameValueMap;
            ruleDefinition = rd;
        }
    }
}
