package securityanalyzer.resourcemodel;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * 
 * Defines an instance of something with parameters. The instance contains the
 * mappings required to resolve the parameters.
 * 
 * @param <T>
 *            The type of thing we are making an instance of
 */
public abstract class FilledParameterContainer<T extends ParameterContainer> {

    /** The thing we are making an instance of */
    protected T base;
    /** These mappings are used to resolve parameters */
    protected Map<ParameterDefinition, String> parameterMappings;

    /**
     * Constructor
     * 
     * @param base
     *            The thing we are making an instance of
     */
    public FilledParameterContainer(T base) {
        this.base = base;
        this.parameterMappings = new HashMap<ParameterDefinition, String>();
    }

    /**
     * Get all mappings that have been defined for this instance
     * 
     * @return
     */
    public Map<ParameterDefinition, String> getAllDefinedMappings() {
        return parameterMappings;
    }

    /**
     * Get all parameters that can be set
     * 
     * @return
     */
    public List<ParameterDefinition> getAllParameters() {
        return base.getParameters();
    }

    /**
     * Get all mappings that have been defined for this instance, combined with
     * mappings that will result from parameters' default values being used
     * 
     * @return
     */
    public Map<ParameterDefinition, String> getAllMappingsResolveDefaults() {
        Map<ParameterDefinition, String> ret = getAllDefinedMappings();
        for (ParameterDefinition pd : base.getParameters()) {
            if ((!ret.keySet().contains(pd)) && (pd.getDefaultVal() != null)) {
                ret.put(pd, pd.getDefaultVal());
            }
        }
        return ret;
    }

    /**
     * Get any parameters that are not mapped to a value, including parameters
     * that have a default value, but do not have a mapping
     * 
     * @return Set of unmapped parameters, or an empty set if there are none
     */
    public Set<ParameterDefinition> getUnmappedParameters() {
        Set<ParameterDefinition> ps = new HashSet<ParameterDefinition>();
        for (ParameterDefinition p : getAllParameters()) {
            if (!parameterMappings.keySet().contains(p)) {
                ps.add(p);
            }
        }
        return ps;
    }

    /**
     * Get any parameters that are not mapped to a value, excluding parameters
     * that have a default value, but do not have a mapping
     * 
     * @return Set of unmapped parameters, or an empty set if there are none
     */
    public Set<ParameterDefinition> getUnmappedParametersWithoutDefaults() {
        Set<ParameterDefinition> ret = new HashSet<ParameterDefinition>();
        for (ParameterDefinition p : getUnmappedParameters()) {
            if (p.getDefaultVal() == null) {
                ret.add(p);
            }
        }
        return ret;
    }

    /**
     * Set a mapping
     * 
     * @param param
     *            the parameter to map
     * @param value
     *            the value to map it to
     * @return previous value associated with the mapping, null if there was no
     *         mapping
     */
    public String setMapping(ParameterDefinition param, String value) {
        return parameterMappings.put(param, value);
    }

    /**
     * get the mapping set for a given parameter
     * 
     * @param param
     *            the parameter
     * @return its mapping
     */
    public String getMapping(ParameterDefinition param) {
        return parameterMappings.get(param);
    }

    /**
     * Get the base object
     * 
     * @return
     */
    public T getBaseObject() {
        return base;
    }

    /**
     * 
     * @param pd
     * @return True if the given parameterdefinition is mapped to its default value, false otherwise
     */
    public boolean isDefault(ParameterDefinition pd) {

        String val = getMapping(pd);
        String def = pd.getDefaultVal();

        if (val == null) {
            return false;
        }

        return val.equals(def);

    }
}
