package securityanalyzer.resourcemodel.reader;

import java.io.File;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;

import securityanalyzer.resourcemodel.ActionDefinition;
import securityanalyzer.resourcemodel.ParameterDefinition;
import securityanalyzer.resourcemodel.ResourceDefinition;
import securityanalyzer.resourcemodel.ResourceModel;
import securityanalyzer.resourcemodel.RuleDefinition;
import securityanalyzer.resourcemodel.schemamodel.ActionParameterType;
import securityanalyzer.resourcemodel.schemamodel.ActionParametersType;
import securityanalyzer.resourcemodel.schemamodel.ActionType;
import securityanalyzer.resourcemodel.schemamodel.AppliedRuleType;
import securityanalyzer.resourcemodel.schemamodel.AppliedRulesType;
import securityanalyzer.resourcemodel.schemamodel.ParameterApplicationType;
import securityanalyzer.resourcemodel.schemamodel.ParameterApplicationsType;
import securityanalyzer.resourcemodel.schemamodel.ParameterType;
import securityanalyzer.resourcemodel.schemamodel.ParametersType;
import securityanalyzer.resourcemodel.schemamodel.ResourceToRulesType;
import securityanalyzer.resourcemodel.schemamodel.ResourceType;
import securityanalyzer.resourcemodel.schemamodel.RuleFieldType;
import securityanalyzer.resourcemodel.schemamodel.SecurityRuleType;

public class ResourceModelReader {

	/**
	 * Uses an instance of the xml schema for defining resource to security rules mappings
	 * to populate the resources and rules of the ResourceModel
	 * @param f The file to use to populate the model
	 */
	public static void setResourcesAndPermissions(File f) {

		ResourceToRulesType rootElement = getRootElement(f);

		if (rootElement == null) {
			System.out.println("Problem getting root element.");
			return;
		}

		addToModelFromRoot(rootElement);

	}

	@SuppressWarnings("unchecked")
	private static ResourceToRulesType getRootElement(File f) {

		Object result;

		try {
                        Class clazz = securityanalyzer.resourcemodel.schemamodel.ObjectFactory.class;
			JAXBContext jc = JAXBContext.newInstance("securityanalyzer.resourcemodel.schemamodel",clazz.getClassLoader());
			Unmarshaller u = jc.createUnmarshaller();
			result = u.unmarshal(f);
		} catch (JAXBException jbe) {
			System.out.println("Error setting up unmarshaller:\n");
			jbe.printStackTrace();
			return null;
		}

		if (result == null) {
			return null;
		}

		if (!(result instanceof JAXBElement)) {
			return null;
		}

		JAXBElement jbe = (JAXBElement) result;

		if (!jbe.getDeclaredType().equals(ResourceToRulesType.class)) {
			return null;
		}

		ResourceToRulesType ret = (ResourceToRulesType) jbe.getValue();

		return ret;

	}

	private static void addToModelFromRoot(ResourceToRulesType rootElement) {

		List<SecurityRuleType> rules = rootElement.getRule();

		addRules(rules);

		List<ResourceType> resources = rootElement.getResource();

		addResources(resources);

		System.out.println();

	}

	private static void addRules(List<SecurityRuleType> rules) {

		for (SecurityRuleType srt : rules) {
			RuleDefinition rd = createRuleDefinition(srt);
			ResourceModel.addRuleDefinition(rd);
		}

	}

	private static RuleDefinition createRuleDefinition(SecurityRuleType srt) {

		RuleDefinition rd = new RuleDefinition(srt.getName(), stripFormatting(srt
				.getDescription()), srt.getType());
		addParametersFromRule(rd, srt.getParameters());
		addAttributesFromRule(rd, srt.getField());

		return rd;

	}

	private static void addParametersFromRule(RuleDefinition rd,
			ParametersType pts) {

		List<ParameterType> lpt = pts.getParameter();

		//TODO: Sort the list by id before adding the parameters
		Collections.sort(lpt, new Comparator<ParameterType>() {
			public int compare(ParameterType pt1, ParameterType pt2) {
				if(pt1.getId() < pt2.getId()) {
					return -1;
				}
				else if(pt1.getId() > pt2.getId()) {
					return 1;
				}
				return 0;
			}
		});

		for (ParameterType pt : pts.getParameter()) {
			ParameterDefinition pad = new ParameterDefinition(pt.getName(),
					"String", pt.getDefault(), stripFormatting(stripFormatting(pt.getDescription())), pt.getId());
			rd.addParameter(pad);
		}

	}
	
	private static void addAttributesFromRule(RuleDefinition rd,
			RuleFieldType ft) {

		// TODO: Right now the cardinality is for only one field per rule.
		// Should fix this to allow any number of fields per rule
		rd.addAttribute(ft.getName().trim(), ft.getValue().trim());

	}

	private static void addResources(List<ResourceType> resources) {

		for (ResourceType rt : resources) {
			ResourceDefinition rd = createResourceDefinition(rt);
			ResourceModel.addResourceDefinition(rd);
		}
	}

	private static ResourceDefinition createResourceDefinition(ResourceType rt) {

		ResourceDefinition rd = new ResourceDefinition(rt.getName(), stripFormatting(rt
				.getDescription()));
		addActionsFromResource(rd, rt.getAction());

		return rd;

	}

	private static void addActionsFromResource(ResourceDefinition rd,
			List<ActionType> actions) {

		for (ActionType at : actions) {
			ActionDefinition ad = createActionDefinition(at, rd);
			rd.addAction(ad);
		}
	}

	private static ActionDefinition createActionDefinition(ActionType at,
			ResourceDefinition rd) {

		ActionDefinition ad = new ActionDefinition(at.getName(), stripFormatting(at
				.getDescription()), rd);

		addActionParameters(ad, at.getActionParameters());
		addAppliedRules(ad, at.getAppliedRules());

		return ad;

	}

	private static void addActionParameters(ActionDefinition ad,
			ActionParametersType apts) {

		// TODO: give actionparameters ids?
		for (ActionParameterType apt : apts.getActionParameter()) {
			ParameterDefinition pd = new ParameterDefinition(apt.getName(), ad
					.getName(), "", stripFormatting(apt.getDescription()), 0);

			ad.addParameter(pd);
                        
                        if(apt.getGroupBy()) {
                            ad.setGroupBy(pd);
                        }
		}

	}

	private static void addAppliedRules(ActionDefinition ad,
			AppliedRulesType prts) {

		for (AppliedRuleType prt : prts.getAppliedRule()) {
			RuleDefinition rd = ResourceModel.getRuleByName(prt.getName());

			// Enforce the mutual relationship between these
			ad.addRequiredPermission(rd);
			rd.addRequiringAction(ad);

			addParameterApplications(ad, rd, prt.getParameterApplications());
		}

	}

	private static void addParameterApplications(ActionDefinition ad,
			RuleDefinition rd, ParameterApplicationsType pats) {

		Map<String, String> nameValueMap = new HashMap<String, String>();
		
		for (ParameterApplicationType pat : pats.getParameterApplication()) {
			nameValueMap.put(pat.getName(), pat.getValue());
		}
		
		ad.addRuleMapping(rd, nameValueMap);

	}
	
	/**
	 * Replaces '\n' with ' ' and '\t' with ""
	 * If s is null, returns null
	 * @param s
	 * @return
	 */
	private static String stripFormatting(String s) {
		
		// There is no description, just give an empty string back
		if(s == null) {
			return null;
		}
		
		s = s.replace("\t", "");
		s = s.replace("\n", " ");
		
		return s;
		
	}



}
